//
// Syd: rock-solid unikernel
// src/syd-path.rs: Write Integrity Force rules for binaries and list executables under PATH
//
// Copyright (c) 2024 Ali Polatel <alip@chesswob.org>
//
// SPDX-License-Identifier: GPL-3.0-or-later

use std::{
    env,
    fs::{canonicalize, read_dir, File},
    io::{BufReader, Write},
    os::unix::ffi::OsStrExt,
    path::{Path, PathBuf},
    process::ExitCode,
};

use getargs::{Opt, Options};
use nix::unistd::{access, AccessFlags};
use rustc_hash::FxHashSet;
use syd::{
    elf::{ElfType, ExecutableFile, LinkingType},
    hash::HashAlgorithm,
};

// Set global allocator to mimalloc if mimalloc feature is enabled (default).
#[cfg(feature = "mimalloc")]
#[global_allocator]
static GLOBAL: mimalloc2_rust::GlobalMiMalloc = mimalloc2_rust::GlobalMiMalloc;

#[derive(Debug, Copy, Clone, PartialEq, Eq, Hash)]
enum Filter {
    ElfFilter32,
    ElfFilter64,
    ElfFilterDynamic,
    ElfFilterStatic,
    ElfFilterPIE,
    ElfFilterNoPIE,
    Script,
}

fn main() -> ExitCode {
    // Configure mimalloc.
    syd::configure_mimalloc();

    #[allow(clippy::disallowed_methods)]
    syd::set_sigpipe_dfl().expect("SIGPIPE");

    let args = env::args().skip(1).collect::<Vec<_>>();
    let mut opts = Options::new(args.iter().map(String::as_str));

    let mut dsyd = false;
    let mut path = env::var("PATH").unwrap_or_default();
    let mut func = HashAlgorithm::Sha512;
    let mut action = "deny";
    let mut limit = 0usize; // 0 means no limit.
    let mut count = 0usize;
    let mut elf_set: FxHashSet<Filter> = FxHashSet::default();

    // SAFETY: We panic on parse errors.
    #[allow(clippy::disallowed_methods)]
    while let Some(opt) = opts.next_opt().expect("next opt") {
        match opt {
            Opt::Short('h') => {
                help();
                return ExitCode::SUCCESS;
            }
            Opt::Short('1') => {
                func = HashAlgorithm::Sha1;
            }
            Opt::Short('2') => {
                func = HashAlgorithm::Sha256;
            }
            Opt::Short('3') => {
                func = HashAlgorithm::Sha384;
            }
            Opt::Short('5') => {
                func = HashAlgorithm::Sha512;
            }
            Opt::Short('m') => {
                func = HashAlgorithm::Md5;
            }
            Opt::Short('d') => {
                action = "deny";
            }
            Opt::Short('k') => {
                action = "kill";
            }
            Opt::Short('w') => {
                action = "warn";
            }
            Opt::Short('p') => {
                path = match opts.value() {
                    Ok(value) => value.to_string(),
                    Err(_) => {
                        eprintln!("-p requires an argument!");
                        return ExitCode::FAILURE;
                    }
                };
            }
            Opt::Short('l') => {
                limit = match opts.value() {
                    Ok(value) => match value.parse::<usize>() {
                        Ok(value) => value,
                        Err(_) => {
                            eprintln!("Invalid argument given to -l!");
                            return ExitCode::FAILURE;
                        }
                    },
                    Err(_) => {
                        eprintln!("-l requires an argument!");
                        return ExitCode::FAILURE;
                    }
                };
            }
            Opt::Short('s') => {
                dsyd = true;
            }
            Opt::Short('e') => match opts.value() {
                Ok(value) => match value {
                    "32" => {
                        if elf_set.contains(&Filter::ElfFilter64) {
                            eprintln!("The option -e32 conflicts with -e64!");
                            return ExitCode::FAILURE;
                        }
                        elf_set.insert(Filter::ElfFilter32);
                    }
                    "64" => {
                        if elf_set.contains(&Filter::ElfFilter32) {
                            eprintln!("The option -e64 conflicts with -e32!");
                            return ExitCode::FAILURE;
                        }
                        elf_set.insert(Filter::ElfFilter64);
                    }
                    "d" => {
                        if elf_set.contains(&Filter::ElfFilterStatic) {
                            eprintln!("The option -ed conflicts with -es!");
                            return ExitCode::FAILURE;
                        }
                        elf_set.insert(Filter::ElfFilterDynamic);
                    }
                    "s" => {
                        if elf_set.contains(&Filter::ElfFilterDynamic) {
                            eprintln!("The option -es conflicts with -ed!");
                            return ExitCode::FAILURE;
                        }
                        elf_set.insert(Filter::ElfFilterStatic);
                    }
                    "p" => {
                        if elf_set.contains(&Filter::ElfFilterNoPIE) {
                            eprintln!("The option -ep conflicts with -eP!");
                            return ExitCode::FAILURE;
                        }
                        elf_set.insert(Filter::ElfFilterPIE);
                    }
                    "P" => {
                        if elf_set.contains(&Filter::ElfFilterPIE) {
                            eprintln!("The option -eP conflicts with -ep!");
                            return ExitCode::FAILURE;
                        }
                        elf_set.insert(Filter::ElfFilterNoPIE);
                    }
                    "x" => {
                        elf_set.insert(Filter::Script);
                    }
                    _ => {
                        eprintln!("Unknown ELF option: -e{value}");
                        return ExitCode::FAILURE;
                    }
                },
                Err(_) => {
                    eprintln!("-e requires an argument!");
                    return ExitCode::FAILURE;
                }
            },
            _ => {
                eprintln!("Unknown option: {opt:?}!");
                return ExitCode::FAILURE;
            }
        }
    }

    let mut path_set: FxHashSet<PathBuf> = FxHashSet::default();
    let dirs = path.split(':');
    for dir in dirs {
        if !Path::new(dir).is_dir() {
            continue;
        }
        if let Ok(entries) = read_dir(dir) {
            for entry in entries.flatten() {
                // Ensure the file is executable
                let path = entry.path();
                if path.is_file() && access(&path, AccessFlags::X_OK).is_ok() {
                    if let Ok(path) = canonicalize(path) {
                        if !path_set.insert(path.clone()) {
                            // Path already seen before.
                            continue;
                        }
                        if !elf_set.is_empty() {
                            // Filter ELF files.
                            #[allow(non_snake_case)]
                            let filter = if let Ok(file) = File::open(&path) {
                                let filter_32 = elf_set.contains(&Filter::ElfFilter32);
                                let filter_64 = elf_set.contains(&Filter::ElfFilter64);
                                let filter_d = elf_set.contains(&Filter::ElfFilterDynamic);
                                let filter_s = elf_set.contains(&Filter::ElfFilterStatic);
                                let filter_p = elf_set.contains(&Filter::ElfFilterPIE);
                                let filter_P = elf_set.contains(&Filter::ElfFilterNoPIE);
                                let filter_x = elf_set.contains(&Filter::Script);
                                let check_linking = filter_d || filter_s || filter_p || filter_P;
                                if let Ok(exe) = ExecutableFile::parse(file, check_linking) {
                                    match exe {
                                        ExecutableFile::Elf {
                                            elf_type: ElfType::Elf32,
                                            ..
                                        } if filter_32 => true,
                                        ExecutableFile::Elf {
                                            elf_type: ElfType::Elf64,
                                            ..
                                        } if filter_64 => true,
                                        ExecutableFile::Elf {
                                            linking_type: Some(LinkingType::Dynamic),
                                            ..
                                        } if filter_d => true,
                                        ExecutableFile::Elf {
                                            linking_type: Some(LinkingType::Static),
                                            ..
                                        } if filter_s => true,
                                        ExecutableFile::Elf { pie: true, .. } if filter_p => true,
                                        ExecutableFile::Elf { pie: false, .. } if filter_P => true,
                                        ExecutableFile::Script if filter_x => true,
                                        _ => false,
                                    }
                                } else {
                                    false
                                }
                            } else {
                                false
                            };

                            #[allow(clippy::disallowed_methods)]
                            if filter {
                                let stdout = std::io::stdout();
                                let mut handle = stdout.lock();
                                handle.write_all(path.as_os_str().as_bytes()).unwrap();
                                handle.write_all(b"\n").unwrap();
                            }
                        } else if let Ok(file) = File::open(&path) {
                            // Write Integrity Force Rules.
                            let reader = BufReader::new(file);
                            if let Ok(key) = syd::hash::hash(reader, func) {
                                let prefix = if dsyd { "/dev/syd/" } else { "" };
                                println!(
                                    "{}force+{}:{}:{}",
                                    prefix,
                                    path.to_string_lossy(),
                                    const_hex::encode(key),
                                    action
                                );
                                if limit > 0 {
                                    count += 1;
                                    if count >= limit {
                                        return ExitCode::SUCCESS;
                                    }
                                }
                            }
                        }
                    }
                }
            }
        }
    }

    ExitCode::SUCCESS
}

fn help() {
    println!("Usage: syd-path [-h12deklpsw]");
    println!("Write Integrity Force rules for binaries under PATH.");
    println!("If at least one of the various *-e* options is specified,");
    println!("List executables with specified information under PATH.");
    println!("Use -m to calculate MD5 instead of SHA3-512 (\x1b[91minsecure\x1b[0m, \x1b[96mPortage\x1b[0m/\x1b[95mPaludis\x1b[0m vdb compat).");
    println!("Use -1 to calculate SHA1 instead of SHA3-512 (\x1b[91minsecure\x1b[0m).");
    println!("Use -2 to calculate SHA3-256 instead of SHA3-512.");
    println!("Use -d for deny, -k for kill, -w for warn.");
    println!("Specify alternative PATH with -p.");
    println!("Use -l <num> to limit by number of entries.");
    println!("Use -s to prefix rules with /dev/syd.");
    println!("Use -e32 to list 32-bit ELF executables under PATH (conflicts with -e64).");
    println!("Use -e64 to list 64-bit ELF executables under PATH (conflicts with -e32).");
    println!("Use -ed to list dynamically linked ELF executables under PATH (conflicts with -es).");
    println!("Use -es to list statically linked ELF executables under PATH (conflicts with -ed).");
    println!("Use -ep to list PIE executables under PATH (conflicts with -eP).");
    println!("Use -eP to list non-PIE executables under PATH (conflicts with -ep).");
    println!("Use -ex to list scripts under PATH.");
}
